import java.util.*;

import jsr.daemon.*;
import jsr.daemon.pausable.*;


/**
 * sample pausable daemon for demonstrating the daemon API
 *
 * After deploying you should start the daemon and read the log output 
 * generated by the daemon for invoking subsequent control actions. If you
 * stress the walker he will take a rest. You can set a maximum number of cycles
 * using the attribute max=<number>.
 *
 * @author   Thomas Kopp
 * @version  0.9, 2000-11-20
 */
public class Walker extends PausableDaemon {

    /**
     * default maximum of walking cycles
     */
    private static final int DEFAULT_MAX = 3;

    /**
     * flag for managing pause state of the walker thread
     */
    private volatile boolean paused = false;

    /**
     * flag for managing activity state of the walker thread
     */
    private volatile boolean ok = true;

    /**
     * maximum number of walking cycles
     */
    private int max = DEFAULT_MAX;
    
    /*
     * walker thread executed by the daemon
     */
    private Thread walker = null;
    
  
    /**
     * method for initializing the walker daemon
     */
    public void init() 
        throws DaemonException {
        log("Hi I am " + getDaemonName() + ". Please wait until I've got my walkin' boots.");

        walker = new Thread() {
                       
                     int n = 0;
                     int m = 0;
                
                     public void run() {
                         log("That's the start of the trip.");
                         while (ok) {
                               n++;
                               try { sleep(1000); }  
                               catch (InterruptedException ex) { interrupt(); ok = false; }
                               if ((paused)||(n >= 10)) {
                                  m++;
                                  if (m >= max) {
                                     log("That's enough for me. I'll stop it."); 
                                     ok = false;
                                  }
                                  else {
                                         if (m > 1) log("... and break number " + m);
                                         else log("Let's take a break.");
                                         pause(new DaemonException("thursty"));
                                         paused = true;
                                         synchronized(this) {
                                             try {  while (paused) wait(); } 
                                             catch (InterruptedException ex) { interrupt(); }
                                             finally { paused = false; }
                                        }
                                        if (ok) log("Ok buddy, we proceed.");
                                        n = 0;
                                  }
                               }
                         }
                         log("We made it.");
                         down(null);
                     }
                 };

        String count = (String)(getAttribute("max"));
        try {
              max = (new Integer(count)).intValue();
              log("The maximum number of cycles is " + max);
              if (max < 1) throw new DaemonException("You decided not to go.");
        }
        catch (NumberFormatException ex) {
              log("We take a well known trail.");
        }
        
        log("Ok, I'm ready. Let's go ...");
    }

    /**
     * method for informing the walker thread
     */
    private void signal() {
        if (paused) {
           paused = false;
           synchronized(walker) {
               walker.notify();
           }
        }
        else walker.start();
    }
    
    /**
     * method for starting the walker thread
     */
    public void start() 
        throws DaemonException {
        log("We start ...");
        signal();
    }
    
    /**
     * method for stopping the walker thread
     */
    public void stop() {
        log("We stop ...");
        paused = true;
    }

    /**
     * method for terminating the walker thread
     */
    public void destroy() {
        log("This is the end.");
        ok = false;
        signal();
    }
}
