package jsr.daemon;

import java.util.*;

/**
 * helper class for dispatching daemon events to bound listeners
 *
 * @author   Thomas Kopp
 * @version  0.9, 2000-11-16
 */
public class DaemonEventHandler {
    
    /**
     * mapping of the listeners bound
     */
    private Set environment = null;  

    /** 
     * default constructor for this event handler
     */
    public DaemonEventHandler() {
        environment = new HashSet();
    }    
    
    /**
     * method for binding a listener to this event handler
     * @param listener the daemon listener to be bound
     */
    public void addListener(DaemonListener listener) {
        environment.add(listener);
    }

    /**
     * method for unbinding a listener from this event handler
     * @param listener the daemon listener to be unbound
     */
    public void removeListener(DaemonListener listener) {
        environment.remove(listener);
    }

    /**
     * method for unbinding all listeners from this event handler
     */
    public void removeListeners() {
        environment.clear();
    }

    /**
     * method for accessing the listeners bound
     * @return iterator of listeners bound to this handler
     */
    protected Iterator getListeners() {
        return environment.iterator();
    }
    
    /**
     * method for informing the event handler about an activity state change
     * @param daemon the daemon changing its activity state
     * @param active the target activity state
     * @param throwable a throwable for indicating a possible exception
     */
    public void active(Daemon daemon, boolean active, Throwable throwable) {
        Iterator listeners = getListeners();
        if (null != listeners) {
           while (listeners.hasNext()) {
                 DaemonListener listener = (DaemonListener)(listeners.next());
                 listener.active(daemon, active, throwable);
           }
        }
    }
}
