package jsr.daemon.pausable;

import jsr.daemon.*;

/**
 * interface supplied by pausable daemons in order to be hosted by a daemon container
 *
 * Pausable daemons may be stopped and restarted during its life cycle.
 *
 * @author   Thomas Kopp
 * @version  0.9, 2000-11-16
 */
public interface Pausable extends Daemon {
    
    /**
     * start method to be called via a daemon control interface
     * method has to be implemented by a pausable daemon in order to be started or restarted after pausing
     * @exception DaemonException thrown by an implementing daemon if starting or restarting fails
     */
    public void start() throws DaemonException;
    
    /**
     * pause method to be called via a daemon control interface
     * method has to be implemented by a pausable daemon in order to be paused if applicable
     */
    public void stop();

    /**
     * pause state indication of a daemon
     * method has to be implemented by a pausable daemon for documenting its current pause state
     * @return true if and only if the implementing pausable daemon is paused
     */
    public boolean isPaused();
}
